/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this file,
 * You can obtain one at http://mozilla.org/MPL/2.0/. */


var EXPORTED_SYMBOLS =
[
    "CascadeMenu",
];

const Cu = Components.utils;

Cu.import("chrome://savefileto-modules/content/common.jsm");
Cu.import("chrome://savefileto-modules/content/Heap.jsm");


function CascadeMenu(config)
{
    this.config = config;
}

CascadeMenu.prototype =
{
    attach: function CM_attach(menu)
    {
        var self = this;
        var doc = menu.ownerDocument;
        
        var menuPopup = this.createElement(doc, "menupopup", {class: this.config.mainMenuClass, position: this.config.mainMenuPosition});
        menuPopup.addEventListener("popupshowing", function(e) { self.onPopupShowing(e); }, false);
        menuPopup.addEventListener("popuphidden", function(e) { self.onPopupHidden(e); }, false);
        menu.appendChild(menuPopup);
        
        menuPopup.addEventListener("command", function(e) { self.onCommand(e); }, false);
    },
    
    onPopupShowing: function CM_onPopupShowing(e)
    {
        e.stopPropagation();
        var menuPopup = e.originalTarget;
        var entry = menuPopup.parentNode.entry;
        
        // main menu
        if (! entry)
        {
            // popup position in ui is determined before popupshowing event
            // so if position has been changed, popup opening must be cancelled,
            // position updated and then menu can be re-opened
            var position = this.config.mainMenuPosition;
            if (position && (position != menuPopup.position))
            {
                e.preventDefault();
                menuPopup.position = position;
                callLater(0, null, function() { menuPopup.parentNode.open = true; });
                return;
            }
            
            this.removeChildren(menuPopup);
            this.populateMainMenu(menuPopup);
        }
        // sub menu
        else
        {
            if (! menuPopup.firstChild)
                this.populateSubMenu(menuPopup, entry);
        }
    },
    
    onPopupHidden: function CM_onPopupHidden(e)
    {
        e.stopPropagation();
        var menuPopup = e.originalTarget;
        var entry = menuPopup.parentNode.entry;
        
        // main menu
        if (! entry)
            this.removeTimer = callLater(0, this, this.removeChildren, [menuPopup]);
    },
    
    onCommand: function CM_onCommand(e)
    {
        var callback = this.config.commandCallback;
        var entry = e.originalTarget.entry;
        callback(e, entry, this);
    },
    
    removeChildren: function CM_removeChildren(element)
    {
        if (this.removeTimer)
        {
            this.removeTimer.cancel();
            this.removeTimer = null;
        }
        
        var child;
        while ((child = element.lastChild) != null)
            element.removeChild(child);
    },
    
    /**
     * Populate main menu from primary data source, 
     * or from fallback data source if primary is empty.
     */
    populateMainMenu: function CM_populateMainMenu(menuPopup)
    {
        let isEmpty = this.populateMenu(menuPopup, this.config.dataSource.entries);
        
        if (isEmpty && this.config.fallbackDataSource)
            this.populateMenu(menuPopup, this.config.fallbackDataSource.entries);
    },
    
    populateSubMenu: function CM_populateSubMenu(menuPopup, entry)
    {
        var children = entry.children;
        
        if (! entry.compacted)
        {
            this.appendMenuItem(menuPopup, entry, this.config.commandLabel);
            
            if (children.length == 0)
                return;
            
            this.appendMenuSeparator(menuPopup);
        }
        
        var minMenuLength = Math.max(2, this.config.minMenuLength);
        var maxMenuLength = this.config.maxMenuLength;
        if (maxMenuLength > 0)
            children = this.compact(children, minMenuLength, maxMenuLength);
        
        for (var child of children)
            this.appendMenuEntry(menuPopup, child);
    },
    
    populateMenu: function CM_populateMenu(menuPopup, entries)
    {
        for (var entry of entries)
            this.appendMenuEntry(menuPopup, entry);
        
        let isEmpty = this.trimMenu(menuPopup);
        return isEmpty;
    },
    
    /**
     * Hides leading, trailing and consecutive <menuseparator> elements from the menu.
     * Returns true if resulting menu is empty.
     */
    trimMenu: function CM_trimMenu(menuPopup)
    {
        let element = menuPopup.firstElementChild;
        let isSeparator = (element && (element.nodeName == "menuseparator"));
        
        // trim leading separators
        while (isSeparator)
        {
            element.hidden = true;
            element = element.nextElementSibling;
            isSeparator = (element && (element.nodeName == "menuseparator"));
        }
        
        let isEmpty = (! element);
        
        // trim consecutive and trailing separators
        while (element)
        {
            let nextElement = element.nextElementSibling;
            let isNextSeparator = ((! nextElement) || (nextElement.nodeName == "menuseparator"));
            
            if (isSeparator && isNextSeparator)
                element.hidden = true;
            
            element = nextElement;
            isSeparator = isNextSeparator;
        }
        
        return isEmpty;
    },
    
    compact: function CM_compact(entries, minLength, maxLength)
    {
        maxLength = Math.max(minLength, maxLength);
        
        if (entries.length <= maxLength)
            return entries;
        
        var prefixEntryList = entries.map(entry => ({prefix: entry.labelLower, entries: [entry]}));
        var result = [prefixEntryList[0]];
        
        // heap for entries common name prefixes (order: longest prefix, alphabetic)
        var reNonAlphanumEnd = /[_\W]+$/;
        var trim = function(s) { return s.replace(reNonAlphanumEnd, ""); };
        var compare = function(a, b) { var d = a.length - b.length; return ((d != 0) ? d : ((b < a) ? -1 : 1)); };
        var prefixHeap = new Heap(compare);
        
        for (var i = 1; i < prefixEntryList.length; i++)
        {
            var prefixEntry = prefixEntryList[i];
            var lastPrefixEntry = result[result.length - 1];
            var prefix = trim(commonPrefix(lastPrefixEntry.prefix, prefixEntry.prefix));
            
            // if last prefixEntry has a sub-list and its prefix equals current, just add the current to the sub-list
            if ((lastPrefixEntry.entries.length > 1) && (prefix == lastPrefixEntry.prefix))
            {
                lastPrefixEntry.entries.push(prefixEntry.entries[0]);
                continue;
            }
            
            result.push(prefixEntry);
            prefixHeap.push(prefix);
            
            // start compacting result list when it grows larger than maxLength
            if (result.length <= maxLength)
                continue;
            
            // compact entries with the longest common prefix into a sub-list, if there is such a prefix
            prefix = prefixHeap.peek();
            if (prefix.length == 0)
                continue;
            
            var resultLength = result.length;
            var prefixLength = prefix.length;
            
            // search entry range that match the longest common prefix
            var lo = 0;
            while ((lo < resultLength) && (result[lo].prefix.substr(0, prefixLength) != prefix))
                lo++;
            var hi = lo + 1;
            while ((hi < resultLength) && (result[hi].prefix.substr(0, prefixLength) == prefix))
                hi++;
            
            // don't allow too aggressive compacting
            var n = hi - lo - 1;
            if ((resultLength - n) < minLength)
                continue;
            
            // combine entries in result[lo,hi[ into a single entry result[lo]
            var subEntries = result[lo].entries;
            var subResult = result.splice(lo + 1, n);
            for (var k = 0; k < subResult.length; k++)
                subEntries = subEntries.concat(subResult[k].entries);
            result[lo] = {prefix: prefix, entries: subEntries};
            
            // remove all instances of the prefix from the heap
            do { prefixHeap.pop(); }
            while ((prefixHeap.length > 0) && (prefixHeap.peek() == prefix))
            
            // add new common prefixes to heap
            if (lo > 0)
                prefixHeap.push(trim(commonPrefix(prefix, result[lo - 1].prefix)));
            if (lo + 1 < result.length)
                prefixHeap.push(trim(commonPrefix(prefix, result[lo + 1].prefix)));
        }
        
        // convert prefixEntries back to normal entries
        entries = [];
        for (var prefixEntry of result)
            if (prefixEntry.entries.length == 1)
                entries.push(prefixEntry.entries[0]);
            else
                entries.push({type: "menu", label: prefixEntry.prefix + "\u2026", hasChildren: true, children: prefixEntry.entries, compacted: true});
        
        return entries;
    },
    
    appendMenuEntry: function CM_appendMenuEntry(menuPopup, entry)
    {
        if (! this.isEntryVisible(menuPopup.ownerDocument, entry))
            return null;
        
        var type = entry.type || "item";
        
        switch (type)
        {
            case "menu":
                if (this.config.trimEmptyMenu && (("hasChildren" in entry) && (! entry.hasChildren)))
                    return this.appendMenuItem(menuPopup, entry);
                else
                    return this.appendMenu(menuPopup, entry);
                
            case "item":
                return this.appendMenuItem(menuPopup, entry);
                
            case "separator":
                return this.appendMenuSeparator(menuPopup);
                
            default:
                Cu.reportError(new Error("invalid entry type: " + type + " " + entry));
                return null;
        }
    },
    
    appendMenu: function CM_appendMenu(menuPopup, entry)
    {
        var self = this;
        var doc = menuPopup.ownerDocument;
        
        var subMenuPopup = this.createElement(doc, "menupopup", {class: this.config.subMenuClass, position: this.getSubMenuPosition(menuPopup)});
        
        var subMenu = this.createElement(doc, "menu", {label: entry.label}, [subMenuPopup]);
        subMenu.entry = entry;
        
        menuPopup.appendChild(subMenu);
        
        return subMenu;
    },
    
    appendMenuItem: function CM_appendMenuItem(menuPopup, entry, label)
    {
        var self = this;
        var doc = menuPopup.ownerDocument;
        
        label = label || entry.label;
        var menuItem = this.createElement(doc, "menuitem", {label: label, command: entry.command});
        menuItem.entry = entry;
        
        menuPopup.appendChild(menuItem);
        
        return menuItem;
    },
    
    appendMenuSeparator: function CM_appendMenuSeparator(menuPopup)
    {
        var doc = menuPopup.ownerDocument;
        
        var menuSeparator = this.createElement(doc, "menuseparator");
        menuPopup.appendChild(menuSeparator);
        
        return menuSeparator;
    },
    
    isEntryVisible: function CM_isEntryVisible(doc, entry)
    {
        if (entry.hidden)
            return false;
        
        // command entry is hidden if there's no corresponding <command> element
        if (("command" in entry) && (! doc.getElementById(entry.command)))
            return false;
        
        return true;
    },
    
    createElement: function CM_createElement(doc, name, attrs, children)
    {
        var element = doc.createElementNS('http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul', name);
        
        if (attrs)
            for (let [attrName, attrValue] of Object.entries(attrs))
                if (attrValue !== undefined)
                    element.setAttribute(attrName, attrValue);
        
        if (children)
            for (var child of children)
                element.appendChild(child);
        
        return element;
    },
    
    alternateMenuPosition:
    {
        start_before: "end_before",
        start_after: "end_after",
        end_before: "start_before",
        end_after: "start_after",
    },
    
    getSubMenuPosition: function(menuPopup)
    {
        if (this.config.alternatingMenuPosition)
            if (menuPopup.position in this.alternateMenuPosition)
                return this.alternateMenuPosition[menuPopup.position];
        
        return this.config.subMenuPosition;
    }
}
